@@ -1,13 +1,12 @@
 /*-
- * Copyright (c) 2003-2004 Tim Kientzle
+ * Copyright (c) 2003-2007 Tim Kientzle
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer
- *    in this position and unchanged.
+ *    notice, this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
@@ -25,21 +24,25 @@
  */
 
 #include "archive_platform.h"
-__FBSDID("$FreeBSD: src/lib/libarchive/archive_read_support_format_iso9660.c,v 1.13 2006/03/21 16:55:46 kientzle Exp $");
-
-#include <sys/stat.h>
+__FBSDID("$FreeBSD: src/lib/libarchive/archive_read_support_format_iso9660.c,v 1.23 2007/05/29 01:00:19 kientzle Exp $");
 
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
 /* #include <stdint.h> */ /* See archive_platform.h */
 #include <stdio.h>
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
 #include <time.h>
-#include <unistd.h>
 
 #include "archive.h"
 #include "archive_entry.h"
 #include "archive_private.h"
+#include "archive_read_private.h"
 #include "archive_string.h"
 
 /*
@@ -70,56 +73,98 @@ __FBSDID("$FreeBSD: src/lib/libarchive/archive_read_support_format_iso9660.c,v 1
  * CDs with a single pass through the data, as required by libarchive.
  */
 
-/* Structure of on-disk PVD. */
-struct iso9660_primary_volume_descriptor {
-	unsigned char	type[1];
-	char	id[5];
-	unsigned char	version[1];
-	char	reserved1[1];
-	char	system_id[32];
-	char	volume_id[32];
-	char	reserved2[8];
-	char	volume_space_size[8];
-	char	reserved3[32];
-	char	volume_set_size[4];
-	char	volume_sequence_number[4];
-	char	logical_block_size[4];
-	char	path_table_size[8];
-	char	type_1_path_table[4];
-	char	opt_type_1_path_table[4];
-	char	type_m_path_table[4];
-	char	opt_type_m_path_table[4];
-	char	root_directory_record[34];
-	char	volume_set_id[128];
-	char	publisher_id[128];
-	char	preparer_id[128];
-	char	application_id[128];
-	char	copyright_file_id[37];
-	char	abstract_file_id[37];
-	char	bibliographic_file_id[37];
-	char	creation_date[17];
-	char	modification_date[17];
-	char	expiration_date[17];
-	char	effective_date[17];
-	char	file_structure_version[1];
-	char	reserved4[1];
-	char	application_data[512];
-};
+/* Structure of on-disk primary volume descriptor. */
+#define PVD_type_offset 0
+#define PVD_type_size 1
+#define PVD_id_offset (PVD_type_offset + PVD_type_size)
+#define PVD_id_size 5
+#define PVD_version_offset (PVD_id_offset + PVD_id_size)
+#define PVD_version_size 1
+#define PVD_reserved1_offset (PVD_version_offset + PVD_version_size)
+#define PVD_reserved1_size 1
+#define PVD_system_id_offset (PVD_reserved1_offset + PVD_reserved1_size)
+#define PVD_system_id_size 32
+#define PVD_volume_id_offset (PVD_system_id_offset + PVD_system_id_size)
+#define PVD_volume_id_size 32
+#define PVD_reserved2_offset (PVD_volume_id_offset + PVD_volume_id_size)
+#define PVD_reserved2_size 8
+#define PVD_volume_space_size_offset (PVD_reserved2_offset + PVD_reserved2_size)
+#define PVD_volume_space_size_size 8
+#define PVD_reserved3_offset (PVD_volume_space_size_offset + PVD_volume_space_size_size)
+#define PVD_reserved3_size 32
+#define PVD_volume_set_size_offset (PVD_reserved3_offset + PVD_reserved3_size)
+#define PVD_volume_set_size_size 4
+#define PVD_volume_sequence_number_offset (PVD_volume_set_size_offset + PVD_volume_set_size_size)
+#define PVD_volume_sequence_number_size 4
+#define PVD_logical_block_size_offset (PVD_volume_sequence_number_offset + PVD_volume_sequence_number_size)
+#define PVD_logical_block_size_size 4
+#define PVD_path_table_size_offset (PVD_logical_block_size_offset + PVD_logical_block_size_size)
+#define PVD_path_table_size_size 8
+#define PVD_type_1_path_table_offset (PVD_path_table_size_offset + PVD_path_table_size_size)
+#define PVD_type_1_path_table_size 4
+#define PVD_opt_type_1_path_table_offset (PVD_type_1_path_table_offset + PVD_type_1_path_table_size)
+#define PVD_opt_type_1_path_table_size 4
+#define PVD_type_m_path_table_offset (PVD_opt_type_1_path_table_offset + PVD_opt_type_1_path_table_size)
+#define PVD_type_m_path_table_size 4
+#define PVD_opt_type_m_path_table_offset (PVD_type_m_path_table_offset + PVD_type_m_path_table_size)
+#define PVD_opt_type_m_path_table_size 4
+#define PVD_root_directory_record_offset (PVD_opt_type_m_path_table_offset + PVD_opt_type_m_path_table_size)
+#define PVD_root_directory_record_size 34
+#define PVD_volume_set_id_offset (PVD_root_directory_record_offset + PVD_root_directory_record_size)
+#define PVD_volume_set_id_size 128
+#define PVD_publisher_id_offset (PVD_volume_set_id_offset + PVD_volume_set_id_size)
+#define PVD_publisher_id_size 128
+#define PVD_preparer_id_offset (PVD_publisher_id_offset + PVD_publisher_id_size)
+#define PVD_preparer_id_size 128
+#define PVD_application_id_offset (PVD_preparer_id_offset + PVD_preparer_id_size)
+#define PVD_application_id_size 128
+#define PVD_copyright_file_id_offset (PVD_application_id_offset + PVD_application_id_size)
+#define PVD_copyright_file_id_size 37
+#define PVD_abstract_file_id_offset (PVD_copyright_file_id_offset + PVD_copyright_file_id_size)
+#define PVD_abstract_file_id_size 37
+#define PVD_bibliographic_file_id_offset (PVD_abstract_file_id_offset + PVD_abstract_file_id_size)
+#define PVD_bibliographic_file_id_size 37
+#define PVD_creation_date_offset (PVD_bibliographic_file_id_offset + PVD_bibliographic_file_id_size)
+#define PVD_creation_date_size 17
+#define PVD_modification_date_offset (PVD_creation_date_offset + PVD_creation_date_size)
+#define PVD_modification_date_size 17
+#define PVD_expiration_date_offset (PVD_modification_date_offset + PVD_modification_date_size)
+#define PVD_expiration_date_size 17
+#define PVD_effective_date_offset (PVD_expiration_date_offset + PVD_expiration_date_size)
+#define PVD_effective_date_size 17
+#define PVD_file_structure_version_offset (PVD_effective_date_offset + PVD_effective_date_size)
+#define PVD_file_structure_version_size 1
+#define PVD_reserved4_offset (PVD_file_structure_version_offset + PVD_file_structure_version_size)
+#define PVD_reserved4_size 1
+#define PVD_application_data_offset (PVD_reserved4_offset + PVD_reserved4_size)
+#define PVD_application_data_size 512
 
 /* Structure of an on-disk directory record. */
-struct iso9660_directory_record {
-	unsigned char length[1];
-	unsigned char ext_attr_length[1];
-	unsigned char extent[8];
-	unsigned char size[8];
-	char date[7];
-	unsigned char flags[1];
-	unsigned char file_unit_size[1];
-	unsigned char interleave[1];
-	unsigned char volume_sequence_number[4];
-	unsigned char name_len[1];
-	char name[1];
-};
+/* Note:  ISO9660 stores each multi-byte integer twice, once in
+ * each byte order.  The sizes here are the size of just one
+ * of the two integers.  (This is why the offset of a field isn't
+ * the same as the offset+size of the previous field.) */
+#define DR_length_offset 0
+#define DR_length_size 1
+#define DR_ext_attr_length_offset 1
+#define DR_ext_attr_length_size 1
+#define DR_extent_offset 2
+#define DR_extent_size 4
+#define DR_size_offset 10
+#define DR_size_size 4
+#define DR_date_offset 18
+#define DR_date_size 7
+#define DR_flags_offset 25
+#define DR_flags_size 1
+#define DR_file_unit_size_offset 26
+#define DR_file_unit_size_size 1
+#define DR_interleave_offset 27
+#define DR_interleave_size 1
+#define DR_volume_sequence_number_offset 28
+#define DR_volume_sequence_number_size 2
+#define DR_name_len_offset 32
+#define DR_name_len_size 1
+#define DR_name_offset 33
 
 /*
  * Our private data.
@@ -167,44 +212,45 @@ struct iso9660 {
 	ssize_t	logical_block_size;
 
 	off_t	entry_sparse_offset;
-	ssize_t	entry_bytes_remaining;
+	int64_t	entry_bytes_remaining;
 };
 
 static void	add_entry(struct iso9660 *iso9660, struct file_info *file);
-static int	archive_read_format_iso9660_bid(struct archive *);
-static int	archive_read_format_iso9660_cleanup(struct archive *);
-static int	archive_read_format_iso9660_read_data(struct archive *,
+static int	archive_read_format_iso9660_bid(struct archive_read *);
+static int	archive_read_format_iso9660_cleanup(struct archive_read *);
+static int	archive_read_format_iso9660_read_data(struct archive_read *,
 		    const void **, size_t *, off_t *);
-static int	archive_read_format_iso9660_read_header(struct archive *,
+static int	archive_read_format_iso9660_read_data_skip(struct archive_read *);
+static int	archive_read_format_iso9660_read_header(struct archive_read *,
 		    struct archive_entry *);
 static const char *build_pathname(struct archive_string *, struct file_info *);
-static void	dump_isodirrec(FILE *, const struct iso9660_directory_record *);
+static void	dump_isodirrec(FILE *, const unsigned char *isodirrec);
 static time_t	time_from_tm(struct tm *);
-static time_t	isodate17(const void *);
-static time_t	isodate7(const void *);
-static int	isPVD(struct iso9660 *, const char *);
+static time_t	isodate17(const unsigned char *);
+static time_t	isodate7(const unsigned char *);
+static int	isPVD(struct iso9660 *, const unsigned char *);
 static struct file_info *next_entry(struct iso9660 *);
-static int	next_entry_seek(struct archive *a, struct iso9660 *iso9660,
+static int	next_entry_seek(struct archive_read *a, struct iso9660 *iso9660,
 		    struct file_info **pfile);
 static struct file_info *
 		parse_file_info(struct iso9660 *iso9660,
-		    struct file_info *parent,
-		    const struct iso9660_directory_record *isodirrec);
+		    struct file_info *parent, const unsigned char *isodirrec);
 static void	parse_rockridge(struct iso9660 *iso9660,
 		    struct file_info *file, const unsigned char *start,
 		    const unsigned char *end);
 static void	release_file(struct iso9660 *, struct file_info *);
-static int	toi(const void *p, int n);
+static unsigned	toi(const void *p, int n);
 
 int
-archive_read_support_format_iso9660(struct archive *a)
+archive_read_support_format_iso9660(struct archive *_a)
 {
+	struct archive_read *a = (struct archive_read *)_a;
 	struct iso9660 *iso9660;
 	int r;
 
-	iso9660 = malloc(sizeof(*iso9660));
+	iso9660 = (struct iso9660 *)malloc(sizeof(*iso9660));
 	if (iso9660 == NULL) {
-		archive_set_error(a, ENOMEM, "Can't allocate iso9660 data");
+		archive_set_error(&a->archive, ENOMEM, "Can't allocate iso9660 data");
 		return (ARCHIVE_FATAL);
 	}
 	memset(iso9660, 0, sizeof(*iso9660));
@@ -216,7 +262,7 @@ archive_read_support_format_iso9660(struct archive *a)
 	    archive_read_format_iso9660_bid,
 	    archive_read_format_iso9660_read_header,
 	    archive_read_format_iso9660_read_data,
-	    NULL,
+	    archive_read_format_iso9660_read_data_skip,
 	    archive_read_format_iso9660_cleanup);
 
 	if (r != ARCHIVE_OK) {
@@ -228,14 +274,14 @@ archive_read_support_format_iso9660(struct archive *a)
 
 
 static int
-archive_read_format_iso9660_bid(struct archive *a)
+archive_read_format_iso9660_bid(struct archive_read *a)
 {
 	struct iso9660 *iso9660;
 	ssize_t bytes_read;
 	const void *h;
-	const char *p;
+	const unsigned char *p;
 
-	iso9660 = *(a->pformat_data);
+	iso9660 = (struct iso9660 *)(a->format->data);
 
 	if (iso9660->bid >= 0)
 		return (iso9660->bid);
@@ -245,10 +291,10 @@ archive_read_format_iso9660_bid(struct archive *a)
 	 * 8 sectors of the volume descriptor table.  Of course,
 	 * if the I/O layer gives us more, we'll take it.
 	 */
-	bytes_read = (a->compression_read_ahead)(a, &h, 32768 + 8*2048);
+	bytes_read = (a->decompressor->read_ahead)(a, &h, 32768 + 8*2048);
 	if (bytes_read < 32768 + 8*2048)
 	    return (iso9660->bid = -1);
-	p = (const char *)h;
+	p = (const unsigned char *)h;
 
 	/* Skip the reserved area. */
 	bytes_read -= 32768;
@@ -259,7 +305,7 @@ archive_read_format_iso9660_bid(struct archive *a)
 		iso9660->bid = isPVD(iso9660, p);
 		if (iso9660->bid > 0)
 			return (iso9660->bid);
-		if (*p == '\xff') /* End-of-volume-descriptor marker. */
+		if (*p == '\177') /* End-of-volume-descriptor marker. */
 			break;
 	}
 
@@ -269,9 +315,8 @@ archive_read_format_iso9660_bid(struct archive *a)
 }
 
 static int
-isPVD(struct iso9660 *iso9660, const char *h)
+isPVD(struct iso9660 *iso9660, const unsigned char *h)
 {
-	const struct iso9660_primary_volume_descriptor *voldesc;
 	struct file_info *file;
 
 	if (h[0] != 1)
@@ -279,32 +324,28 @@ isPVD(struct iso9660 *iso9660, const char *h)
 	if (memcmp(h+1, "CD001", 5) != 0)
 		return (0);
 
-
-	voldesc = (const struct iso9660_primary_volume_descriptor *)h;
-	iso9660->logical_block_size = toi(&voldesc->logical_block_size, 2);
+	iso9660->logical_block_size = toi(h + PVD_logical_block_size_offset, 2);
 
 	/* Store the root directory in the pending list. */
-	file = parse_file_info(iso9660, NULL,
-	    (struct iso9660_directory_record *)&voldesc->root_directory_record);
+	file = parse_file_info(iso9660, NULL, h + PVD_root_directory_record_offset);
 	add_entry(iso9660, file);
 	return (48);
 }
 
 static int
-archive_read_format_iso9660_read_header(struct archive *a,
+archive_read_format_iso9660_read_header(struct archive_read *a,
     struct archive_entry *entry)
 {
-	struct stat st;
 	struct iso9660 *iso9660;
 	struct file_info *file;
 	ssize_t bytes_read;
 	int r;
 
-	iso9660 = *(a->pformat_data);
+	iso9660 = (struct iso9660 *)(a->format->data);
 
-	if (!a->archive_format) {
-		a->archive_format = ARCHIVE_FORMAT_ISO9660;
-		a->archive_format_name = "ISO9660";
+	if (!a->archive.archive_format) {
+		a->archive.archive_format = ARCHIVE_FORMAT_ISO9660;
+		a->archive.archive_format_name = "ISO9660";
 	}
 
 	/* Get the next entry that appears after the current offset. */
@@ -316,22 +357,20 @@ archive_read_format_iso9660_read_header(struct archive *a,
 	iso9660->entry_sparse_offset = 0; /* Offset for sparse-file-aware clients. */
 
 	/* Set up the entry structure with information about this entry. */
-	memset(&st, 0, sizeof(st));
-	st.st_mode = file->mode;
-	st.st_uid = file->uid;
-	st.st_gid = file->gid;
-	st.st_nlink = file->nlinks;
-	st.st_ino = file->inode;
-	st.st_mtime = file->mtime;
-	st.st_ctime = file->ctime;
-	st.st_atime = file->atime;
-	st.st_size = iso9660->entry_bytes_remaining;
-	archive_entry_copy_stat(entry, &st);
+	archive_entry_set_mode(entry, file->mode);
+	archive_entry_set_uid(entry, file->uid);
+	archive_entry_set_gid(entry, file->gid);
+	archive_entry_set_nlink(entry, file->nlinks);
+	archive_entry_set_ino(entry, file->inode);
+	archive_entry_set_mtime(entry, file->mtime, 0);
+	archive_entry_set_ctime(entry, file->ctime, 0);
+	archive_entry_set_atime(entry, file->atime, 0);
+	archive_entry_set_size(entry, iso9660->entry_bytes_remaining);
 	archive_string_empty(&iso9660->pathname);
 	archive_entry_set_pathname(entry,
 	    build_pathname(&iso9660->pathname, file));
 	if (file->symlink.s != NULL)
-		archive_entry_set_symlink(entry, file->symlink.s);
+		archive_entry_copy_symlink(entry, file->symlink.s);
 
 	/* If this entry points to the same data as the previous
 	 * entry, convert this into a hardlink to that entry.
@@ -350,7 +389,7 @@ archive_read_format_iso9660_read_header(struct archive *a,
 	/* If the offset is before our current position, we can't
 	 * seek backwards to extract it, so issue a warning. */
 	if (file->offset < iso9660->current_position) {
-		archive_set_error(a, ARCHIVE_ERRNO_MISC,
+		archive_set_error(&a->archive, ARCHIVE_ERRNO_MISC,
 		    "Ignoring out-of-order file");
 		iso9660->entry_bytes_remaining = 0;
 		iso9660->entry_sparse_offset = 0;
@@ -363,46 +402,44 @@ archive_read_format_iso9660_read_header(struct archive *a,
 	archive_strcpy(&iso9660->previous_pathname, iso9660->pathname.s);
 
 	/* If this is a directory, read in all of the entries right now. */
-	if (S_ISDIR(st.st_mode)) {
+	if (archive_entry_filetype(entry) == AE_IFDIR) {
 		while (iso9660->entry_bytes_remaining > 0) {
 			const void *block;
 			const unsigned char *p;
 			ssize_t step = iso9660->logical_block_size;
 			if (step > iso9660->entry_bytes_remaining)
 				step = iso9660->entry_bytes_remaining;
-			bytes_read = (a->compression_read_ahead)(a, &block, step);
+			bytes_read = (a->decompressor->read_ahead)(a, &block, step);
 			if (bytes_read < step) {
-				archive_set_error(a, ARCHIVE_ERRNO_MISC,
+				archive_set_error(&a->archive, ARCHIVE_ERRNO_MISC,
 	    "Failed to read full block when scanning ISO9660 directory list");
 				release_file(iso9660, file);
 				return (ARCHIVE_FATAL);
 			}
 			if (bytes_read > step)
 				bytes_read = step;
-			(a->compression_read_consume)(a, bytes_read);
+			(a->decompressor->consume)(a, bytes_read);
 			iso9660->current_position += bytes_read;
 			iso9660->entry_bytes_remaining -= bytes_read;
-			for (p = block;
+			for (p = (const unsigned char *)block;
 			     *p != 0 && p < (const unsigned char *)block + bytes_read;
 			     p += *p) {
-				const struct iso9660_directory_record *dr
-				    = (const struct iso9660_directory_record *)p;
 				struct file_info *child;
 
 				/* Skip '.' entry. */
-				if (dr->name_len[0] == 1
-				    && dr->name[0] == '\0')
+				if (*(p + DR_name_len_offset) == 1
+				    && *(p + DR_name_offset) == '\0')
 					continue;
 				/* Skip '..' entry. */
-				if (dr->name_len[0] == 1
-				    && dr->name[0] == '\001')
+				if (*(p + DR_name_len_offset) == 1
+				    && *(p + DR_name_offset) == '\001')
 					continue;
-				child = parse_file_info(iso9660, file, dr);
+				child = parse_file_info(iso9660, file, p);
 				add_entry(iso9660, child);
 				if (iso9660->seenRockridge) {
-					a->archive_format =
+					a->archive.archive_format =
 					    ARCHIVE_FORMAT_ISO9660_ROCKRIDGE;
-					a->archive_format_name =
+					a->archive.archive_format_name =
 					    "ISO9660 with Rockridge extensions";
 				}
 			}
@@ -414,13 +451,22 @@ archive_read_format_iso9660_read_header(struct archive *a,
 }
 
 static int
-archive_read_format_iso9660_read_data(struct archive *a,
+archive_read_format_iso9660_read_data_skip(struct archive_read *a)
+{
+	/* Because read_next_header always does an explicit skip
+	 * to the next entry, we don't need to do anything here. */
+	(void)a; /* UNUSED */
+	return (ARCHIVE_OK);
+}
+
+static int
+archive_read_format_iso9660_read_data(struct archive_read *a,
     const void **buff, size_t *size, off_t *offset)
 {
 	ssize_t bytes_read;
 	struct iso9660 *iso9660;
 
-	iso9660 = *(a->pformat_data);
+	iso9660 = (struct iso9660 *)(a->format->data);
 	if (iso9660->entry_bytes_remaining <= 0) {
 		*buff = NULL;
 		*size = 0;
@@ -428,9 +474,9 @@ archive_read_format_iso9660_read_data(struct archive *a,
 		return (ARCHIVE_EOF);
 	}
 
-	bytes_read = (a->compression_read_ahead)(a, buff, 1);
+	bytes_read = (a->decompressor->read_ahead)(a, buff, 1);
 	if (bytes_read == 0)
-		archive_set_error(a, ARCHIVE_ERRNO_MISC,
+		archive_set_error(&a->archive, ARCHIVE_ERRNO_MISC,
 		    "Truncated input file");
 	if (bytes_read <= 0)
 		return (ARCHIVE_FATAL);
@@ -441,23 +487,25 @@ archive_read_format_iso9660_read_data(struct archive *a,
 	iso9660->entry_sparse_offset += bytes_read;
 	iso9660->entry_bytes_remaining -= bytes_read;
 	iso9660->current_position += bytes_read;
-	(a->compression_read_consume)(a, bytes_read);
+	(a->decompressor->consume)(a, bytes_read);
 	return (ARCHIVE_OK);
 }
 
 static int
-archive_read_format_iso9660_cleanup(struct archive *a)
+archive_read_format_iso9660_cleanup(struct archive_read *a)
 {
 	struct iso9660 *iso9660;
 	struct file_info *file;
 
-	iso9660 = *(a->pformat_data);
+	iso9660 = (struct iso9660 *)(a->format->data);
 	while ((file = next_entry(iso9660)) != NULL)
 		release_file(iso9660, file);
 	archive_string_free(&iso9660->pathname);
 	archive_string_free(&iso9660->previous_pathname);
+	if (iso9660->pending_files)
+		free(iso9660->pending_files);
 	free(iso9660);
-	*(a->pformat_data) = NULL;
+	(a->format->data) = NULL;
 	return (ARCHIVE_OK);
 }
 
@@ -467,68 +515,72 @@ archive_read_format_iso9660_cleanup(struct archive *a)
  */
 static struct file_info *
 parse_file_info(struct iso9660 *iso9660, struct file_info *parent,
-    const struct iso9660_directory_record *isodirrec)
+    const unsigned char *isodirrec)
 {
 	struct file_info *file;
+	size_t name_len;
+	int flags;
 
 	/* TODO: Sanity check that name_len doesn't exceed length, etc. */
 
 	/* Create a new file entry and copy data from the ISO dir record. */
-	file = malloc(sizeof(*file));
+	file = (struct file_info *)malloc(sizeof(*file));
 	if (file == NULL)
 		return (NULL);
 	memset(file, 0, sizeof(*file));
 	file->parent = parent;
 	if (parent != NULL)
 		parent->refcount++;
-	file->offset = toi(isodirrec->extent, 4)
+	file->offset = toi(isodirrec + DR_extent_offset, DR_extent_size)
 	    * iso9660->logical_block_size;
-	file->size = toi(isodirrec->size, 4);
-	file->mtime = isodate7(isodirrec->date);
+	file->size = toi(isodirrec + DR_size_offset, DR_size_size);
+	file->mtime = isodate7(isodirrec + DR_date_offset);
 	file->ctime = file->atime = file->mtime;
-	file->name = malloc(isodirrec->name_len[0] + 1);
+	name_len = (size_t)*(const unsigned char *)(isodirrec + DR_name_len_offset);
+	file->name = (char *)malloc(name_len + 1);
 	if (file->name == NULL) {
 		free(file);
 		return (NULL);
 	}
-	memcpy(file->name, isodirrec->name, isodirrec->name_len[0]);
-	file->name[(int)isodirrec->name_len[0]] = '\0';
-	if (isodirrec->flags[0] & 0x02)
-		file->mode = S_IFDIR | 0700;
+	memcpy(file->name, isodirrec + DR_name_offset, name_len);
+	file->name[name_len] = '\0';
+	flags = *(isodirrec + DR_flags_offset);
+	if (flags & 0x02)
+		file->mode = AE_IFDIR | 0700;
 	else
-		file->mode = S_IFREG | 0400;
+		file->mode = AE_IFREG | 0400;
 
 	/* Rockridge extensions overwrite information from above. */
 	{
 		const unsigned char *rr_start, *rr_end;
 		rr_end = (const unsigned char *)isodirrec
-		    + isodirrec->length[0];
-		rr_start = (const unsigned char *)isodirrec->name
-		    + isodirrec->name_len[0];
-		if ((isodirrec->name_len[0] & 1) == 0)
+		    + *(isodirrec + DR_length_offset);
+		rr_start = (const unsigned char *)(isodirrec + DR_name_offset
+		    + name_len);
+		if ((name_len & 1) == 0)
 			rr_start++;
 		rr_start += iso9660->suspOffset;
 		parse_rockridge(iso9660, file, rr_start, rr_end);
 	}
 
 	/* DEBUGGING: Warn about attributes I don't yet fully support. */
-	if ((isodirrec->flags[0] & ~0x02) != 0) {
+	if ((flags & ~0x02) != 0) {
 		fprintf(stderr, "\n ** Unrecognized flag: ");
 		dump_isodirrec(stderr, isodirrec);
 		fprintf(stderr, "\n");
-	} else if (toi(isodirrec->volume_sequence_number, 2) != 1) {
+	} else if (toi(isodirrec + DR_volume_sequence_number_offset, 2) != 1) {
 		fprintf(stderr, "\n ** Unrecognized sequence number: ");
 		dump_isodirrec(stderr, isodirrec);
 		fprintf(stderr, "\n");
-	} else if (isodirrec->file_unit_size[0] != 0) {
+	} else if (*(isodirrec + DR_file_unit_size_offset) != 0) {
 		fprintf(stderr, "\n ** Unexpected file unit size: ");
 		dump_isodirrec(stderr, isodirrec);
 		fprintf(stderr, "\n");
-	} else if (isodirrec->interleave[0] != 0) {
+	} else if (*(isodirrec + DR_interleave_offset) != 0) {
 		fprintf(stderr, "\n ** Unexpected interleave: ");
 		dump_isodirrec(stderr, isodirrec);
 		fprintf(stderr, "\n");
-	} else if (isodirrec->ext_attr_length[0] != 0) {
+	} else if (*(isodirrec + DR_ext_attr_length_offset) != 0) {
 		fprintf(stderr, "\n ** Unexpected extended attribute length: ");
 		dump_isodirrec(stderr, isodirrec);
 		fprintf(stderr, "\n");
@@ -547,7 +599,7 @@ add_entry(struct iso9660 *iso9660, struct file_info *file)
 
 		if (new_size < 1024)
 			new_size = 1024;
-		new_pending_files = malloc(new_size * sizeof(new_pending_files[0]));
+		new_pending_files = (struct file_info **)malloc(new_size * sizeof(new_pending_files[0]));
 		if (new_pending_files == NULL)
 			__archive_errx(1, "Out of memory");
 		memcpy(new_pending_files, iso9660->pending_files,
@@ -610,7 +662,7 @@ parse_rockridge(struct iso9660 *iso9660, struct file_info *file,
 
 				data++;  /* Skip flag byte. */
 				data_length--;
-				file->name = malloc(data_length + 1);
+				file->name = (char *)malloc(data_length + 1);
 				if (file->name != NULL) {
 					free(old_name);
 					memcpy(file->name, data, data_length);
@@ -831,7 +883,7 @@ release_file(struct iso9660 *iso9660, struct file_info *file)
 }
 
 static int
-next_entry_seek(struct archive *a, struct iso9660 *iso9660,
+next_entry_seek(struct archive_read *a, struct iso9660 *iso9660,
     struct file_info **pfile)
 {
 	struct file_info *file;
@@ -857,22 +909,13 @@ fprintf(stderr, " *** Discarding CE data.\n");
 			offset = file->offset;
 
 		/* Seek forward to the start of the entry. */
-		while (iso9660->current_position < offset) {
-			ssize_t step = offset - iso9660->current_position;
-			ssize_t bytes_read;
-			const void *buff;
-
-			if (step > iso9660->logical_block_size)
-				step = iso9660->logical_block_size;
-			bytes_read = (a->compression_read_ahead)(a, &buff, step);
-			if (bytes_read <= 0) {
-				release_file(iso9660, file);
-				return (ARCHIVE_FATAL);
-			}
-			if (bytes_read > step)
-				bytes_read = step;
-			iso9660->current_position += bytes_read;
-			(a->compression_read_consume)(a, bytes_read);
+		if (iso9660->current_position < offset) {
+			off_t step = offset - iso9660->current_position;
+			off_t bytes_read;
+			bytes_read = (a->decompressor->skip)(a, step);
+			if (bytes_read < 0)
+				return (bytes_read);
+			iso9660->current_position = offset;
 		}
 
 		/* We found body of file; handle it now. */
@@ -888,13 +931,13 @@ fprintf(stderr, " *** Discarding CE data.\n");
 
 			file->ce_offset = 0;
 			file->ce_size = 0;
-			bytes_read = (a->compression_read_ahead)(a, &p, size);
+			bytes_read = (a->decompressor->read_ahead)(a, &p, size);
 			if (bytes_read > size)
 				bytes_read = size;
 			rr_start = (const unsigned char *)p;
 			parse_rockridge(iso9660, file, rr_start,
 			    rr_start + bytes_read);
-			(a->compression_read_consume)(a, bytes_read);
+			(a->decompressor->consume)(a, bytes_read);
 			iso9660->current_position += bytes_read;
 			add_entry(iso9660, file);
 		}
@@ -937,7 +980,7 @@ next_entry(struct iso9660 *iso9660)
 	return (r);
 }
 
-static int
+static unsigned int
 toi(const void *p, int n)
 {
 	const unsigned char *v = (const unsigned char *)p;
@@ -949,10 +992,9 @@ toi(const void *p, int n)
 }
 
 static time_t
-isodate7(const void *p)
+isodate7(const unsigned char *v)
 {
 	struct tm tm;
-	const unsigned char *v = (const unsigned char *)p;
 	int offset;
 	memset(&tm, 0, sizeof(tm));
 	tm.tm_year = v[0];
@@ -961,8 +1003,8 @@ isodate7(const void *p)
 	tm.tm_hour = v[3];
 	tm.tm_min = v[4];
 	tm.tm_sec = v[5];
-	/* v[6] is the timezone offset, in 1/4-hour increments. */
-	offset = ((const signed char *)p)[6];
+	/* v[6] is the signed timezone offset, in 1/4-hour increments. */
+	offset = ((const signed char *)v)[6];
 	if (offset > -48 && offset < 52) {
 		tm.tm_hour -= offset / 4;
 		tm.tm_min -= (offset % 4) * 15;
@@ -971,10 +1013,9 @@ isodate7(const void *p)
 }
 
 static time_t
-isodate17(const void *p)
+isodate17(const unsigned char *v)
 {
 	struct tm tm;
-	const unsigned char *v = (const unsigned char *)p;
 	int offset;
 	memset(&tm, 0, sizeof(tm));
 	tm.tm_year = (v[0] - '0') * 1000 + (v[1] - '0') * 100
@@ -985,8 +1026,8 @@ isodate17(const void *p)
 	tm.tm_hour = (v[8] - '0') * 10 + (v[9] - '0');
 	tm.tm_min = (v[10] - '0') * 10 + (v[11] - '0');
 	tm.tm_sec = (v[12] - '0') * 10 + (v[13] - '0');
-	/* v[16] is the timezone offset, in 1/4-hour increments. */
-	offset = ((const signed char *)p)[16];
+	/* v[16] is the signed timezone offset, in 1/4-hour increments. */
+	offset = ((const signed char *)v)[16];
 	if (offset > -48 && offset < 52) {
 		tm.tm_hour -= offset / 4;
 		tm.tm_min -= (offset % 4) * 15;
@@ -1004,30 +1045,37 @@ time_from_tm(struct tm *t)
 {
 #if HAVE_TIMEGM
 	return (timegm(t));
-#else
+#elif HAVE_STRUCT_TM_TM_GMTOFF
 	/*
 	 * Unfortunately, timegm() isn't standard.  The standard
 	 * mktime() function is a close match, except that it uses
-	 * local timezone instead of GMT.  Close enough for now.
-	 * Note that it is not possible to emulate timegm() using
-	 * standard interfaces:
-	 *   * ANSI C90 does not even guarantee that time_t is
-	 *     an arithmetic type, so time adjustments can only be
-	 *     done by manipulating struct tm elements.  You cannot
-	 *     portably calculate time_t values.
-	 *   * POSIX does promise that time_t is an arithmetic type
-	 *     measured in seconds, so you can do time_t calculations
-	 *     while remaining POSIX-compliant.
-	 *   * Neither ANSI nor POSIX provides an easy way to measure
-	 *     the timezone offset, so you can't adjust mktime() to
-	 *     work like timegm().
-	 *   * POSIX does not promise that the epoch begins in 1970,
-	 *     so you can't write a portable timegm() function from
-	 *     scratch.
+	 * local timezone instead of GMT.  You can compensate for
+	 * this by adding the timezone and DST offsets back in, at
+	 * the cost of two calls to mktime().
 	 */
-	time_t result = mktime(t);
-	/* TODO: Find a way to improve this approximation to timegm(). */
-	return result;
+	mktime(t); /* Normalize the time and get the TZ offset. */
+	t->tm_sec += t->tm_gmtoff; /* Try to adjust for the timezone and DST.*/
+	if (t->tm_isdst)
+		t->tm_hour -= 1;
+	return (mktime(t)); /* Re-convert. */
+#else
+	/*
+	 * If you don't have tm_gmtoff, let's try resetting the timezone
+	 * (yecch!).
+	 */
+	time_t ret;
+	char *tz;
+
+	tz = getenv("TZ");
+	setenv("TZ", "UTC 0", 1);
+	tzset();
+	ret = mktime(t);
+	if (tz)
+	    setenv("TZ", tz, 1);
+	else
+	    unsetenv("TZ");
+	tzset();
+	return ret;
 #endif
 }
 
@@ -1046,16 +1094,26 @@ build_pathname(struct archive_string *as, struct file_info *file)
 }
 
 static void
-dump_isodirrec(FILE *out, const struct iso9660_directory_record *isodirrec)
+dump_isodirrec(FILE *out, const unsigned char *isodirrec)
 {
-	fprintf(out, " l %d,", isodirrec->length[0]);
-	fprintf(out, " a %d,", isodirrec->ext_attr_length[0]);
-	fprintf(out, " ext 0x%x,", toi(isodirrec->extent, 4));
-	fprintf(out, " s %d,", toi(isodirrec->size, 4));
-	fprintf(out, " f 0x%02x,", isodirrec->flags[0]);
-	fprintf(out, " u %d,", isodirrec->file_unit_size[0]);
-	fprintf(out, " ilv %d,", isodirrec->interleave[0]);
-	fprintf(out, " seq %d,", toi(isodirrec->volume_sequence_number,2));
-	fprintf(out, " nl %d:", isodirrec->name_len[0]);
-	fprintf(out, " `%.*s'", isodirrec->name_len[0], isodirrec->name);
+	fprintf(out, " l %d,",
+	    toi(isodirrec + DR_length_offset, DR_length_size));
+	fprintf(out, " a %d,",
+	    toi(isodirrec + DR_ext_attr_length_offset, DR_ext_attr_length_size));
+	fprintf(out, " ext 0x%x,",
+	    toi(isodirrec + DR_extent_offset, DR_extent_size));
+	fprintf(out, " s %d,",
+	    toi(isodirrec + DR_size_offset, DR_extent_size));
+	fprintf(out, " f 0x%02x,",
+	    toi(isodirrec + DR_flags_offset, DR_flags_size));
+	fprintf(out, " u %d,",
+	    toi(isodirrec + DR_file_unit_size_offset, DR_file_unit_size_size));
+	fprintf(out, " ilv %d,",
+	    toi(isodirrec + DR_interleave_offset, DR_interleave_size));
+	fprintf(out, " seq %d,",
+	    toi(isodirrec + DR_volume_sequence_number_offset, DR_volume_sequence_number_size));
+	fprintf(out, " nl %d:",
+	    toi(isodirrec + DR_name_len_offset, DR_name_len_size));
+	fprintf(out, " `%.*s'",
+	    toi(isodirrec + DR_name_len_offset, DR_name_len_size), isodirrec + DR_name_offset);
 }
