@@ -1,13 +1,12 @@
 /*-
- * Copyright (c) 2003-2004 Tim Kientzle
+ * Copyright (c) 2003-2007 Tim Kientzle
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer
- *    in this position and unchanged.
+ *    notice, this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
@@ -25,23 +24,30 @@
  */
 
 #include "archive_platform.h"
-__FBSDID("$FreeBSD: src/lib/libarchive/archive_write_set_format_cpio.c,v 1.6 2005/09/21 04:25:05 kientzle Exp $");
+__FBSDID("$FreeBSD: src/lib/libarchive/archive_write_set_format_cpio.c,v 1.11 2007/05/29 01:00:19 kientzle Exp $");
 
-#include <sys/stat.h>
+#ifdef HAVE_ERRNO_H
 #include <errno.h>
+#endif
 #include <stdio.h>
+#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
+#endif
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
 
 #include "archive.h"
 #include "archive_entry.h"
 #include "archive_private.h"
+#include "archive_write_private.h"
 
-static ssize_t	archive_write_cpio_data(struct archive *, const void *buff,
-		    size_t s);
-static int	archive_write_cpio_finish(struct archive *);
-static int	archive_write_cpio_finish_entry(struct archive *);
-static int	archive_write_cpio_header(struct archive *,
+static ssize_t	archive_write_cpio_data(struct archive_write *,
+		    const void *buff, size_t s);
+static int	archive_write_cpio_finish(struct archive_write *);
+static int	archive_write_cpio_destroy(struct archive_write *);
+static int	archive_write_cpio_finish_entry(struct archive_write *);
+static int	archive_write_cpio_header(struct archive_write *,
 		    struct archive_entry *);
 static int	format_octal(int64_t, void *, int);
 static int64_t	format_octal_recursive(int64_t, char *, int);
@@ -68,17 +74,18 @@ struct cpio_header {
  * Set output format to 'cpio' format.
  */
 int
-archive_write_set_format_cpio(struct archive *a)
+archive_write_set_format_cpio(struct archive *_a)
 {
+	struct archive_write *a = (struct archive_write *)_a;
 	struct cpio *cpio;
 
 	/* If someone else was already registered, unregister them. */
-	if (a->format_finish != NULL)
-		(a->format_finish)(a);
+	if (a->format_destroy != NULL)
+		(a->format_destroy)(a);
 
-	cpio = malloc(sizeof(*cpio));
+	cpio = (struct cpio *)malloc(sizeof(*cpio));
 	if (cpio == NULL) {
-		archive_set_error(a, ENOMEM, "Can't allocate cpio data");
+		archive_set_error(&a->archive, ENOMEM, "Can't allocate cpio data");
 		return (ARCHIVE_FATAL);
 	}
 	memset(cpio, 0, sizeof(*cpio));
@@ -89,50 +96,50 @@ archive_write_set_format_cpio(struct archive *a)
 	a->format_write_data = archive_write_cpio_data;
 	a->format_finish_entry = archive_write_cpio_finish_entry;
 	a->format_finish = archive_write_cpio_finish;
+	a->format_destroy = archive_write_cpio_destroy;
 	a->archive_format = ARCHIVE_FORMAT_CPIO_POSIX;
 	a->archive_format_name = "POSIX cpio";
 	return (ARCHIVE_OK);
 }
 
 static int
-archive_write_cpio_header(struct archive *a, struct archive_entry *entry)
+archive_write_cpio_header(struct archive_write *a, struct archive_entry *entry)
 {
 	struct cpio *cpio;
 	const char *p, *path;
 	int pathlength, ret;
-	const struct stat	*st;
 	struct cpio_header	 h;
 
-	cpio = a->format_data;
+	cpio = (struct cpio *)a->format_data;
 	ret = 0;
 
 	path = archive_entry_pathname(entry);
 	pathlength = strlen(path) + 1; /* Include trailing null. */
-	st = archive_entry_stat(entry);
 
 	memset(&h, 0, sizeof(h));
 	format_octal(070707, &h.c_magic, sizeof(h.c_magic));
-	format_octal(st->st_dev, &h.c_dev, sizeof(h.c_dev));
+	format_octal(archive_entry_dev(entry), &h.c_dev, sizeof(h.c_dev));
 	/*
 	 * TODO: Generate artificial inode numbers rather than just
 	 * re-using the ones off the disk.  That way, the 18-bit c_ino
 	 * field only limits the number of files in the archive.
 	 */
-	if (st->st_ino > 0777777) {
-		archive_set_error(a, ERANGE, "large inode number truncated");
+	if (archive_entry_ino(entry) > 0777777) {
+		archive_set_error(&a->archive, ERANGE, "large inode number truncated");
 		ret = ARCHIVE_WARN;
 	}
 
-	format_octal(st->st_ino & 0777777, &h.c_ino, sizeof(h.c_ino));
-	format_octal(st->st_mode, &h.c_mode, sizeof(h.c_mode));
-	format_octal(st->st_uid, &h.c_uid, sizeof(h.c_uid));
-	format_octal(st->st_gid, &h.c_gid, sizeof(h.c_gid));
-	format_octal(st->st_nlink, &h.c_nlink, sizeof(h.c_nlink));
-	if (S_ISBLK(st->st_mode) || S_ISCHR(st->st_mode))
-	    format_octal(st->st_rdev, &h.c_rdev, sizeof(h.c_rdev));
+	format_octal(archive_entry_ino(entry) & 0777777, &h.c_ino, sizeof(h.c_ino));
+	format_octal(archive_entry_mode(entry), &h.c_mode, sizeof(h.c_mode));
+	format_octal(archive_entry_uid(entry), &h.c_uid, sizeof(h.c_uid));
+	format_octal(archive_entry_gid(entry), &h.c_gid, sizeof(h.c_gid));
+	format_octal(archive_entry_nlink(entry), &h.c_nlink, sizeof(h.c_nlink));
+	if (archive_entry_filetype(entry) == AE_IFBLK
+	    || archive_entry_filetype(entry) == AE_IFCHR)
+	    format_octal(archive_entry_dev(entry), &h.c_rdev, sizeof(h.c_rdev));
 	else
 	    format_octal(0, &h.c_rdev, sizeof(h.c_rdev));
-	format_octal(st->st_mtime, &h.c_mtime, sizeof(h.c_mtime));
+	format_octal(archive_entry_mtime(entry), &h.c_mtime, sizeof(h.c_mtime));
 	format_octal(pathlength, &h.c_namesize, sizeof(h.c_namesize));
 
 	/* Symlinks get the link written as the body of the entry. */
@@ -140,36 +147,36 @@ archive_write_cpio_header(struct archive *a, struct archive_entry *entry)
 	if (p != NULL  &&  *p != '\0')
 		format_octal(strlen(p), &h.c_filesize, sizeof(h.c_filesize));
 	else
-		format_octal(st->st_size, &h.c_filesize, sizeof(h.c_filesize));
+		format_octal(archive_entry_size(entry), &h.c_filesize, sizeof(h.c_filesize));
 
-	ret = (a->compression_write)(a, &h, sizeof(h));
+	ret = (a->compressor.write)(a, &h, sizeof(h));
 	if (ret != ARCHIVE_OK)
 		return (ARCHIVE_FATAL);
 
-	ret = (a->compression_write)(a, path, pathlength);
+	ret = (a->compressor.write)(a, path, pathlength);
 	if (ret != ARCHIVE_OK)
 		return (ARCHIVE_FATAL);
 
-	cpio->entry_bytes_remaining = st->st_size;
+	cpio->entry_bytes_remaining = archive_entry_size(entry);
 
 	/* Write the symlink now. */
 	if (p != NULL  &&  *p != '\0')
-		ret = (a->compression_write)(a, p, strlen(p));
+		ret = (a->compressor.write)(a, p, strlen(p));
 
 	return (ret);
 }
 
 static ssize_t
-archive_write_cpio_data(struct archive *a, const void *buff, size_t s)
+archive_write_cpio_data(struct archive_write *a, const void *buff, size_t s)
 {
 	struct cpio *cpio;
 	int ret;
 
-	cpio = a->format_data;
+	cpio = (struct cpio *)a->format_data;
 	if (s > cpio->entry_bytes_remaining)
 		s = cpio->entry_bytes_remaining;
 
-	ret = (a->compression_write)(a, buff, s);
+	ret = (a->compressor.write)(a, buff, s);
 	cpio->entry_bytes_remaining -= s;
 	if (ret >= 0)
 		return (s);
@@ -188,10 +195,10 @@ format_octal(int64_t v, void *p, int digits)
 
 	max = (((int64_t)1) << (digits * 3)) - 1;
 	if (v >= 0  &&  v <= max) {
-	    format_octal_recursive(v, p, digits);
+	    format_octal_recursive(v, (char *)p, digits);
 	    ret = 0;
 	} else {
-	    format_octal_recursive(max, p, digits);
+	    format_octal_recursive(max, (char *)p, digits);
 	    ret = -1;
 	}
 	return (ret);
@@ -208,39 +215,44 @@ format_octal_recursive(int64_t v, char *p, int s)
 }
 
 static int
-archive_write_cpio_finish(struct archive *a)
+archive_write_cpio_finish(struct archive_write *a)
 {
 	struct cpio *cpio;
-	struct stat st;
 	int er;
 	struct archive_entry *trailer;
 
-	cpio = a->format_data;
+	cpio = (struct cpio *)a->format_data;
 	trailer = archive_entry_new();
-	memset(&st, 0, sizeof(st));
-	st.st_nlink = 1;
-	archive_entry_copy_stat(trailer, &st);
+	archive_entry_set_nlink(trailer, 1);
 	archive_entry_set_pathname(trailer, "TRAILER!!!");
 	er = archive_write_cpio_header(a, trailer);
 	archive_entry_free(trailer);
-
-	free(cpio);
-	a->format_data = NULL;
 	return (er);
 }
 
 static int
-archive_write_cpio_finish_entry(struct archive *a)
+archive_write_cpio_destroy(struct archive_write *a)
+{
+	struct cpio *cpio;
+
+	cpio = (struct cpio *)a->format_data;
+	free(cpio);
+	a->format_data = NULL;
+	return (ARCHIVE_OK);
+}
+
+static int
+archive_write_cpio_finish_entry(struct archive_write *a)
 {
 	struct cpio *cpio;
 	int to_write, ret;
 
-	cpio = a->format_data;
+	cpio = (struct cpio *)a->format_data;
 	ret = ARCHIVE_OK;
 	while (cpio->entry_bytes_remaining > 0) {
 		to_write = cpio->entry_bytes_remaining < a->null_length ?
 		    cpio->entry_bytes_remaining : a->null_length;
-		ret = (a->compression_write)(a, a->nulls, to_write);
+		ret = (a->compressor.write)(a, a->nulls, to_write);
 		if (ret != ARCHIVE_OK)
 			return (ret);
 		cpio->entry_bytes_remaining -= to_write;
